import networkx as nx
import numpy as np

word_delimiters = ['!', '"', '#', '%', '&', '(', ')', '*', '+',
                   ',', '-', '.', '...', '......', '/', ':', ';',
                   '<', '=', '>', '?', '@', '[', ']', '^', '_',
                   '`', '{', '|', '}', '~', '，', '。', '。。。',
                   '。。。。。。', '！', '；', '？', '：', '、', '（', '）',
                   '\t', '\n', '”', '’' , '“', '‘', "'"]

sentence_delimiters = ['？', '?', '；', ';', '！', '!', '。', '……', '…', '\n']

allow_pos = ['an', 'nz', 'vn', 'v', 'vd', 'x', 'n', 'nt', 'nz', 'd']

def get_stop_words(path_stop_words):
    """

    Parameters:
    ----------
    path_stop_words: 停用词文件路径, 其中一行存储一个停用词

    Returns:
    -------
    stop_words: list, 停用词列表

    """
    stop_words = set()
    if path_stop_words:
        with open(path_stop_words, 'r') as fr:
            stop_words = set([word.strip().replace('\n', '') for word in fr.readlines()])

    return stop_words

def get_similarity(words1, words2):
    """
    计算句子之间的相似度

    公式: similarity = |A∩B| / (log(|A|) + log(|B|))

    Parameters:
    ----------
    words1: list, 词列表

    words2: list, 词列表

    Returns:
    -------
    sim_value: float, 句子之间的相似度
    """

    if len(words1) == 0 or len(words2) == 0:
        sim_value = 0
    else:
        sim_value = float(len(set(words1).intersection(set(words2))) / (np.log(len(words1)) + np.log(len(words2))))

    return sim_value


def sort_sentences(sentences, words,
                   sim_func=get_similarity,
                   pagerank_config=None,
                   pr_error_handle='iterator'):
    f"""
    基于TextRank方法对句子以及分词结果进行提取摘要

    Parameters:
    ----------
    sentences: list, 切分后的结果句子列表

    words: list, 切分后的结果词列表

    pagerank_config: dict, networkx.pagerank的参数字典

    pr_error_handle: str, pagerank不收敛的时候的处理策略, iterator增加迭代轮次（兜底）, tolerance增加迭代轮次前后的差值阈值

    Returns:
    -------
    list_res: list, 结果列表
    """
    list_res = []

    # 默认的PR收敛时的参数
    pr_alpha = 1
    pr_max_iter = 200
    pr_tol = 1e-6

    if pagerank_config is None:
        pagerank_config = {'alpha': pr_alpha,
                           'max_iter': pr_max_iter,
                           'tol': pr_tol}

    len_sentences = len(sentences)

    # 初始化句子之间的无向权图, 整体为N*N的矩阵
    graph = np.zeros((len_sentences, len_sentences))

    # 计算权重, 权重由切词的相似度进行计算, 由于是无向的, a(ij) = a(ji)
    for i in range(len_sentences):
        for j in range(len_sentences):
            sim_value = sim_func(words[i], words[j])
            graph[i, j] = sim_value
            graph[j, i] = sim_value

    # 构造无向权图
    nx_graph = nx.from_numpy_matrix(graph)

    # 计算PR值, 注意, 初始参数在计算PR值时可能不收敛, 这个时候可以
    flag = True

    while flag:
        # noinspection PyBroadException
        try:
            ## 开始计算PR值, 可能存在不收敛的情况
            pr_values = nx.pagerank(nx_graph, **pagerank_config)
            ## 成功收敛则停止循环
            flag = False

        except Exception:
            ## 如果PR不收敛, 以提升迭代前后轮次之间的差值为策略
            if pr_error_handle == 'tolerance':
                pr_tol = min(pr_tol * 10, 0)
            ## 否则以提升迭代轮次作为策略
            else:
                pr_max_iter += 50

            pagerank_config = {'alpha': pr_alpha,
                               'max_iter': pr_max_iter,
                               'tol': pr_tol}

    # pr_values: 一个dict, {index:pr, index:pr}
    for idx, val in sorted(pr_values.items(), key=lambda x: x[1], reverse=True):
        list_res.append({'sentence': sentences[idx], 'weight': val})

    return list_res




import jieba.posseg as pseg
import re
from . import utils

class WordSegment(object):
    """
    用于切词的类, 因为基于jieba分词, 所以需要提前load_userdict加载自定义词库

    Attributes:
    ----------
    stop_words: list, 停用词列表

    allow_pos: list, 选用的词性列表

    punct: list, 剔除的特殊字符列表
    """

    def __init__(self, path_stop_words=None, allow_pos=None, punct=utils.word_delimiters):
        self.stop_words = utils.get_stop_words(path_stop_words=path_stop_words)
        self.allow_pos = allow_pos
        self.punct = punct

    def segment(self, content, use_stop_words=False, use_allow_pos=False, use_is_lower=True):
        """
        对单个句子进行分词, 其中包含四种分词结果（用于后面抽取摘要的相似值计算）

        Parameters:
        ----------
        content: str, 输入的文本, 可包含特殊字符

        Returns:
        -------
        list_words: list, 分词结果列表
        """
        # 先剔除空格符
        content = content.replace(" ", "")

        # 转化大小写
        if use_is_lower:
            content = content.lower()

        # pseg切出词和词性, 其中剔除特殊字符
        pseg_res = [item for item in pseg.lcut(content) if item.word not in self.punct and item.word]

        # 剔除停用词, 这里停用词可能是空集
        if use_stop_words:
            pseg_res = [item for item in pseg_res if item.word not in self.stop_words]

        # 剔除词性, None的时候则保留全词, 否则筛选对应的词性
        if use_allow_pos:
            pseg_res = [item for item in pseg_res if item.flag in self.allow_pos]

        list_words = [item.word for item in pseg_res]

        return list_words

    def seg_sentences(self, sentences, use_stop_words=False, use_allow_pos=False, use_is_lower=True):
        """
        对多个句子进行分词

        Parameters:
        ----------
        sentences: list, 一个列表含一个字符串句子

        Returns:
        -------
        list_sents: 二维list, 一个列表含一个句子的分词结果列表,
        """

        list_sents = [self.segment(sent,
                                   use_stop_words=use_stop_words,
                                   use_allow_pos=use_allow_pos,
                                   use_is_lower=use_is_lower) for sent in sentences]

        return list_sents

class SentenceSegment(object):
    """
    切句的类
    这里的切句是指根据特殊符号将多个句子进行拆分, 不是将句子拆成词
    e.g.
    AAAA。BBBB。 -> [AAAA, BBBB]

    Attributes:
    ----------
    delimeters: list, 切句使用的特殊字符列表
    """
    def __init__(self, delimiters=utils.sentence_delimiters):
        self.delimeters = delimiters

    def segment(self, content, use_is_lower=True):
        """
        根据特殊符号将多个句子进行拆分

        e.g.
        AAAA。BBBB。 -> [AAAA, BBBB]

        Parameters:
        ----------
        content: str, 输入的语料, 字符串长串

        delemeters: list, 切割用的字符串

        use_is_lower: bool, 是否要转化为小写

        Returns:
        -------
        res_split: list, 切分后的句子列表
        """
        # 先剔除空格符
        content = content.replace(" ", "")

        # 转化大小写
        if use_is_lower:
            content = content.lower()

        # 基于分隔符进行split
        pattern = '[{0}*]'.format('|'.join(self.delimeters))
        ## 注意可能存在AAAA;;;;!BBBB这种连续特殊字符的场景, split后会生成空字符串需要剔除
        res_split = [s for s in re.split(pattern, content) if len(s) > 0]

        return res_split







from TextRank4Sentence import TextRank4Sentence
from segment import WordSegment
import os


if __name__ == '__main__':
    work_path = os.getcwd()
    path_stop_words = os.path.join(work_path, 'stop_words.txt')
    corpus = " 近日，云南省德宏州瑞丽市出现本土新冠肺炎疫情，瑞丽市疫情防控工作指挥部已经发布通告，从7月5日8时起，所有人员非必要不进出瑞丽；自7月6日12时起，将瑞丽市姐告国门社区调整为中风险地区，其他区域为低风险地区。鉴于云南省德宏州疫情变化形势，为严格落实“外防输入、内’防反弹”的防控策略，有效控制‘’和降''低疫情传播风险，市疾控中心向广大市民发出紧急提醒"

    t4ks = TextRank4Sentence()
    t4ks.analyze(text=corpus, source='no_filters')

    ws = WordSegment()
    ws.segment(content=corpus)
    print(ws.segment(content=corpus))




from . import utils
from .segment import WordSegment, SentenceSegment

class TextRank4Sentence(object):
    """
    提取关键句

    Attributes:
    ----------
    sentences: list, 切分后的结果句子列表

    allow_pos: list, 切词时保留的词性

    path_stop_words: str, 停用词表路径

    punct: list, 剔除的特殊字符列表

    delimiters: list, 切句使用的特殊字符列表

    ws: class object, 用于切词的类

    ss: class object, 用于切句的类

    words_no_filters: list, 分词结果, 不剔除停用词、不剔除词性、

    words_stop_words_filters: list, 分词结果, 剔除停用词、不剔除词性

    words_flags_filters: list, 分词结果, 不剔除停用词, 剔除词性

    words_all_filters: list, 分词结果, 剔除停用词、剔除词性

    key_sentences: list, 提取的全量句子以及其权重(PR值)
    """
    def __init__(self):
        self.sentences = None
        self.allow_pos = None
        self.path_stop_words = None
        self.punct = None
        self.delimiters  = None
        self.ws = None
        self.ss = None

        self.words_no_filters = None
        self.words_stop_words_filters = None
        self.words_flags_filters = None
        self.words_all_filters = None

        self.key_sentences = None

    def analyze(self, text,
                punct=utils.word_delimiters,
                delimiters=utils.sentence_delimiters,
                source='all_filters',
                sim_func=utils.get_similarity,
                path_stop_words=None,
                allow_pos=utils.allow_pos,
                is_lower=True):
        """
        基于TextRank提取全量的句子以及其权重(PR值)


        Parameters:
        ----------
        text: str, 输入的文本

        punct: list, 剔除的特殊字符列表

        delimiters: list, 切句使用的特殊字符列表

        source: str, 最后用于提取摘要句计算PR时用的分词结果

        sim_func: function, 用于计算两个词列表的相似度函数, 支持用户自定义该函数, 格式为 func(list, list) -> float

        path_stop_words: str, 停用词表路径

        allow_pos: list, 切词时保留的词性

        is_lower: bool, 是否要转小写字母

        Returns:
        -------


        """
        self.punct = punct
        self.delimiters = delimiters
        self.allow_pos = allow_pos
        self.path_stop_words = path_stop_words

        self.ws = WordSegment(path_stop_words=self.path_stop_words,
                              allow_pos=self.allow_pos,
                              punct=self.punct)

        self.ss = SentenceSegment(delimiters=delimiters)

        # 切分句子, 这里会输出一个列表, 一个列表包含一个句子
        # PS:这里也处理了大小写转化
        self.sentences = self.ss.segment(content=text, use_is_lower=is_lower)

        # 对所有的句子开始切词, 其中要区分四种情况
        ## 1.no_filters, 不剔除停用词、不剔除词性、
        self.words_no_filters = self.ws.seg_sentences(sentences=self.sentences,
                                                      use_stop_words=False,
                                                      use_allow_pos=False)
        ## 2.stop_words_filters, 剔除停用词、不剔除词性
        self.words_stop_words_filters = self.ws.seg_sentences(sentences=self.sentences,
                                                              use_stop_words=True,
                                                              use_allow_pos=False)

        ## 3.flags_filters, 不剔除停用词, 剔除词性
        self.words_flags_filters = self.ws.seg_sentences(sentences=self.sentences,
                                                         use_stop_words=True,
                                                         use_allow_pos=False)

        ## 4.all_filters, 剔除停用词、剔除词性
        self.words_all_filters = self.ws.seg_sentences(sentences=self.sentences,
                                                       use_stop_words=True,
                                                       use_allow_pos=True)

        # 基于source决定输入的分词列表
        if source == 'no_filters':
            _words = self.words_no_filters
        elif source == 'stop_words_filters':
            _words = self.words_stop_words_filters
        elif source == 'flags_filters':
            _words = self.words_flags_filters
        else:
            ## all_filters作为兜底
            _words = self.words_all_filters

        self.key_sentences = utils.sort_sentences(sentences=self.sentences,
                                                  words=_words,
                                                  sim_func=sim_func)

    def get_key_sentences(self, top_k=5, sentences_min_len=5, with_weight=True):
        """
        筛选需要的topK摘要句

        Parameters:
        ----------
        top_k: int, 获取topN权重的句子

        sentences_min_len: int, 获取大于等于长度阈值的句子

        with_weight: bool, 是否要带着权重(PR值)进行输出

        Returns:
        -------
        摘要句列表

        """
        list_res = []

        sentences = [item['sentence'] for item in self.key_sentences if len(item['sentence']) >= sentences_min_len]
        weights = [item['weight'] for item in self.key_sentences if len(item['sentence']) >= sentences_min_len]

        list_res.append(sentences)

        if with_weight:
            list_res.append(weights)

        return list(zip(*list_res))[:top_k]











# -*- coding: utf-8 -*-
from setuptools import setup

LONGDOC = """
我家还蛮大的, 欢迎你们来我家van.

https://github.com/skykiseki

textrank4ch
====

"基于textrank的关键字摘取与摘要句摘取", 


具体textrank的知识参考, 我个人觉得还不错:     
[《PageRank 笔记》](https://clvsit.blog.csdn.net/article/details/90322422)
[《textrank提取关键词与关键句》](https://blog.csdn.net/weixin_40746796/article/details/89963268)


完整文档见 ``README.md``

GitHub: https://github.com/skykiseki/textrank4ch
"""

setup(name='textrank4ch',
      version='1.1.2',
      description='Key words extracting and Key sentences extracting',
      long_description=LONGDOC,
      long_description_content_type="text/markdown",
      author='Wei, Zhihui',
      author_email='evelinesdd@qq.com',
      url='https://github.com/skykiseki/textrank4ch',
      license="MIT",
      classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: Chinese (Traditional)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Indexing',
        'Topic :: Text Processing :: Linguistic',
      ],
      python_requires='>=3.6',
      install_requires=[
        'pandas',
        'numpy',
        'jieba',
        'networkx'
      ],
      keywords='NLP,Chinese,Chinese words,textrank',
      packages=['textrank4ch'],
      package_dir={'textrank4ch':'textrank4ch'}
)


import jieba.analyse
import jieba
from . import utils

class TextRank4Keywords(object):
    """
    提取关键字

    这里直接改造成引用jieba的textrank结果, 不再有逻辑分岔, 且集成了停用词路径等参数

    Attributes:
    ----------
    path_stop_words: str, 停用词表路径

    text: str, 文本字符串

    keywords: list, 关键字列表, 一个元素包含关键词(必选), 词性(可选), 权重(可选)

    allow_pos: list, 切词时保留的词性
    """
    def __init__(self):
        self.path_stop_words = None
        self.text = None
        self.keywords = None
        self.allow_pos = None

    def analyze(self, text, path_stop_words=None, allow_pos=utils.allow_pos, is_lower=True):
        """
        对文本进行分析, 找出所有的关键词候选列表

        Parameters:
        ----------
        text: str, 文本字符串

        path_stop_words: str, 停用词表路径

        is_lower: bool, 是否要转小写字母

        Returns:
        -------
        self

        """
        self.path_stop_words = path_stop_words
        self.text = text.replace(" ", "")
        self.allow_pos = allow_pos

        if is_lower:
            self.text = self.text.lower()

        if path_stop_words:
            jieba.analyse.set_stop_words(stop_words_path=path_stop_words)

        self.keywords = jieba.analyse.textrank(sentence=self.text,
                                               topK=None,
                                               allowPOS=self.allow_pos,
                                               withWeight=True,
                                               withFlag=True)

    def get_key_words(self, top_k=6, word_min_len=1, with_weight=True, with_flag=False):
        """
        获取需要的关键词

        Parameters:
        ----------
        top_k: int, 获取topN权重的词

        word_min_len: int, 获取大于等于长度阈值的关键词

        with_weight: bool, 是否要带着权重(PR值)进行输出

        with_flag: bool, 是否要带着词性进行输出

        Returns:
        -------
        关键词列表

        """
        list_res = []

        words = [item[0].word for item in self.keywords if len(item[0].word) >= word_min_len]
        weights = [item[1] for item in self.keywords if len(item[0].word) >= word_min_len]
        flags = [item[0].flag for item in self.keywords if len(item[0].word) >= word_min_len]

        list_res.append(words)

        if with_weight:
            list_res.append(weights)

        if with_flag:
            list_res.append(flags)

        return list(zip(*list_res))[:top_k]

